/***************************************************************************

    osd_tool.c

    OS-dependent code interface for tools

    Copyright (c) 1996-2006, Nicola Salmoria and the MAME Team.
    Visit http://mamedev.org for licensing and usage restrictions.

***************************************************************************/

#define _LARGEFILE64_SOURCE
#define _XOPEN_SOURCE 500

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "osd_cpu.h"
#include "osd_tool.h"

struct _osd_tool_file {
	int fd;
};

/*-------------------------------------------------
    is_physical_drive - clue to Win32 code that
    we're reading a physical drive directly
-------------------------------------------------*/

int osd_is_physical_drive(const char *file)
{
#if defined(SDLMAME_MACOSX) || defined(SDLMAME_WIN32) || defined(SDLMAME_NO64BITIO)
	struct stat st;
	if(stat(file, &st))
		return 0;
#else
	struct stat64 st;
	if(stat64(file, &st))
		return 0;
#endif
	return S_ISBLK(st.st_mode);
}



/*-------------------------------------------------
    osd_get_physical_drive_geometry - retrieves
    geometry for physical drives
-------------------------------------------------*/

int osd_get_physical_drive_geometry(const char *filename, UINT32 *cylinders, UINT32 *heads, UINT32 *sectors, UINT32 *bps)
{
	fprintf(stderr, "Can't get geometry, sorry\n");
	exit(1);
}



/*-------------------------------------------------
    osd_get_file_size - returns the 64-bit file size
    for a file
-------------------------------------------------*/

UINT64 osd_get_file_size(const char *file)
{
#if defined(SDLMAME_MACOSX) || defined(SDLMAME_WIN32) || defined(SDLMAME_NO64BITIO)
	struct stat st;
	if(stat(file, &st))
		return 0;
#else
	struct stat64 st;
	if(stat64(file, &st))
		return 0;
#endif
	return st.st_size;
}



osd_tool_file *osd_tool_fopen(const char *filename, const char *mode)
{
	osd_tool_file *file = malloc(sizeof(osd_tool_file));
	int omode;

	/* convert the mode into opening mode */
	if (!strcmp(mode, "rb"))
		omode = O_RDONLY;
	else if (!strcmp(mode, "rb+"))
		omode = O_RDWR;
	else if (!strcmp(mode, "wb"))
		omode = O_RDWR|O_CREAT|O_TRUNC;
	else
		return NULL;

	/* if this is a physical drive, don't creat/trunc */
	if (osd_is_physical_drive(filename) && (omode & O_TRUNC))
		omode = O_RDWR;

#if defined(SDLMAME_MACOSX) || defined(SDLMAME_WIN32) || defined(SDLMAME_NO64BITIO)
	file->fd = open(filename, omode, 0666);
#else
	file->fd = open64(filename, omode, 0666);
#endif
	if (file->fd < 0) {
		free(file);
		return NULL;
	}
	return file;
}



void osd_tool_fclose(osd_tool_file *file)
{
	close(file->fd);
	free(file);
}



UINT32 osd_tool_fread(osd_tool_file *file, UINT64 offset, UINT32 count, void *buffer)
{
#ifdef SDLMAME_MACOSX
	UINT32 res = pread(file->fd, buffer, count, offset);
#elif defined(SDLMAME_LINUX)
	UINT32 res = pread64(file->fd, buffer, count, offset);
#elif defined(SDLMAME_WIN32) || defined(SDLMAME_NO64BITIO)
	UINT32 res;
	lseek(file->fd, offset, SEEK_SET);
	res = read(file->fd, buffer, count); 
#else
#error Unknown SDL SUBARCH!
#endif
	if(res == (UINT32)-1)
		res = 0;
	return res;
}



UINT32 osd_tool_fwrite(osd_tool_file *file, UINT64 offset, UINT32 count, const void *buffer)
{
#ifdef SDLMAME_MACOSX
	UINT32 res = pwrite(file->fd, buffer, count, offset);
#elif defined(SDLMAME_LINUX)
	UINT32 res = pwrite64(file->fd, buffer, count, offset);
#elif defined(SDLMAME_WIN32) || defined(SDLMAME_NO64BITIO)
	UINT32 res;
	lseek(file->fd, offset, SEEK_SET);
	res = write(file->fd, buffer, count); 
#else
#error Unknown SDL SUBARCH!
#endif
	if(res == (UINT32)-1)
		res = 0;
	return res;
}



UINT64 osd_tool_flength(osd_tool_file *file)
{
#if defined(SDLMAME_MACOSX) || defined(SDLMAME_WIN32) || defined(SDLMAME_NO64BITIO)
	UINT64 res = lseek(file->fd, 0, SEEK_END);
#else
	UINT64 res = lseek64(file->fd, 0, SEEK_END);
#endif
	if(res < 0)
		res = 0;
	return res;
}


