//============================================================
//
//  sound.c - SDL implementation of MAME sound routines
//
//  Copyright (c) 1996-2006, Nicola Salmoria and the MAME Team.
//  Visit http://mamedev.org for licensing and usage restrictions.
//
//  SDLMAME by Olivier Galibert and R. Belmont
//
//============================================================

// standard sdl header
#include <SDL/SDL.h>
#include <unistd.h>

#ifdef SDLMAME_WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

// MAME headers
#include "driver.h"
#include "osdepend.h"
#include "sound/wavwrite.h"
#include "window.h"
#include "video.h"

//============================================================
//	IMPORTS
//============================================================

extern int verbose;


//============================================================
//	DEBUGGING
//============================================================

#define LOG_SOUND		0
#define DISPLAY_UNDEROVERFLOW	0



//============================================================
//	PARAMETERS
//============================================================

#define INGORE_UNDERFLOW_FRAMES	100

// the local buffer is what the stream buffer feeds from
// note that this needs to be large enough to buffer at frameskip 11
// for 30fps games like Tapper; we will scale the value down based
// on the actual framerate of the game
#define MAX_BUFFER_SIZE			(128 * 1024)

// this is the maximum number of extra samples we will ask for
// per frame (I know this looks like a lot, but most of the
// time it will generally be nowhere close to this)
#define MAX_SAMPLE_ADJUST		32

// number of samples per SDL callback
#define SDL_XFER_SAMPLES		1024

// maximum audio latency
#define MAX_AUDIO_LATENCY		10

//============================================================
//	GLOBAL VARIABLES
//============================================================

// global parameters
int 					attenuation = 0;

int							audio_latency;
char *						wavwrite;

//============================================================
//	LOCAL VARIABLES
//============================================================

static int				initialized_audio;
static int				buf_locked;

static INT8				*stream_buffer;
static volatile INT32	  		stream_playpos;

static UINT32				stream_buffer_size;
static UINT32				stream_buffer_in;
static volatile UINT32	 		stream_write_cursor;
static UINT32				nBlockAlign;

// buffer over/underflow counts
static int				total_frames;
static int				buffer_underflows;
static int				buffer_overflows;

// global sample tracking
static double				samples_per_frame;
static double				samples_left_over;
static UINT32				samples_this_frame;

// sample rate adjustments
static int				current_adjustment = 0;
static int				lower_thresh;
static int				upper_thresh;

// debugging
#if LOG_SOUND
static FILE *				sound_log1;
#define sound_log stderr
#endif

static void *				wavptr;

// sound enable
static int snd_enabled;

//============================================================
//	PROTOTYPES
//============================================================

static int			sdl_init(void);
static void			sdl_kill(void);
static int			sdl_create_buffers(void);
static void			sdl_destroy_buffers(void);



//============================================================
//	bytes_in_stream_buffer
//============================================================

INLINE int bytes_in_stream_buffer(void)
{
	if (stream_buffer_in > stream_playpos)
	{
		#if LOG_SOUND
		fprintf(sound_log, "1 in: %d playpos %d\n", stream_buffer_in, stream_playpos);
		#endif
		return stream_buffer_in - stream_playpos;
	}
	else
	{
		#if LOG_SOUND
		fprintf(sound_log, "2 in: %d playpos %d\n", stream_buffer_in, stream_playpos);
		#endif
		return stream_buffer_size + stream_buffer_in - stream_playpos;
	}

	return 0;
}



//============================================================
//	osd_start_audio_stream
//============================================================
int osd_start_audio_stream(int stereo)
{
#if LOG_SOUND
	sound_log = fopen("sound.log", "w");
#endif

	// skip if sound disabled
	if (Machine->sample_rate != 0)
	{
		// attempt to initialize SDL
		if (sdl_init())
			return 1;

		// set the startup volume
		osd_set_mastervolume(attenuation);
	}

	// determine the number of samples per frame
	samples_per_frame = (double)Machine->sample_rate / (double)Machine->screen[0].refresh;

	// compute how many samples to generate the first frame
	samples_left_over = samples_per_frame;
	samples_this_frame = (UINT32)samples_left_over;
	samples_left_over -= (double)samples_this_frame;

	// create wav file
	if( wavwrite != NULL )
	{
		wavptr = wav_open( wavwrite, Machine->sample_rate, 2);
	}
	else
	{
		wavptr = NULL;
	}

	// return the samples to play the first frame
	return samples_this_frame;
}



//============================================================
//	osd_stop_audio_stream
//============================================================

void osd_stop_audio_stream(void)
{
	if( wavptr != NULL )
	{
		wav_close( wavptr );
		wavptr = NULL;
	}

	// if nothing to do, don't do it
	if (Machine->sample_rate == 0)
		return;
	
	// kill the buffers and dsound
	sdl_kill();
	sdl_destroy_buffers();
	
	// print out over/underflow stats
	if (verbose && (buffer_overflows || buffer_underflows))
		fprintf(stderr, "Sound buffer: overflows=%d underflows=%d\n", buffer_overflows, buffer_underflows);

#if LOG_SOUND
	if (sound_log)
		fprintf(sound_log, "Sound buffer: overflows=%d underflows=%d\n", buffer_overflows, buffer_underflows);
	fclose(sound_log);
#endif
}


//============================================================
//	update_sample_adjustment
//============================================================

static void update_sample_adjustment(int buffered)
{
	static int consecutive_lows = 0;
	static int consecutive_mids = 0;
	static int consecutive_highs = 0;

	// if we're not throttled don't bother
	if (!video_config.throttle)
	{
#if LOG_SOUND
	fprintf(sound_log, "unthrottled, no adjustment\n");
#endif
		consecutive_lows = 0;
		consecutive_mids = 0;
		consecutive_highs = 0;
		current_adjustment = 0;
		return;
	}

#if LOG_SOUND
	fprintf(sound_log, "update_sample_adjustment: %d\n", buffered);
#endif

	// do we have too few samples in the buffer?
	if (buffered < lower_thresh)
	{
		// keep track of how many consecutive times we get this condition
		consecutive_lows++;
		consecutive_mids = 0;
		consecutive_highs = 0;

		// adjust so that we generate more samples per frame to compensate
//		current_adjustment = (consecutive_lows < MAX_SAMPLE_ADJUST) ? consecutive_lows : MAX_SAMPLE_ADJUST;
		current_adjustment = ((lower_thresh - buffered) < MAX_SAMPLE_ADJUST) ? (lower_thresh-buffered) : MAX_SAMPLE_ADJUST;

#if LOG_SOUND
		fprintf(sound_log, "  (too low - adjusting to %d)\n", current_adjustment);
#endif
	}

	// do we have too many samples in the buffer?
	else if (buffered > upper_thresh)
	{
		// keep track of how many consecutive times we get this condition
		consecutive_lows = 0;
		consecutive_mids = 0;
		consecutive_highs++;

		// adjust so that we generate fewer samples per frame to compensate
//		current_adjustment = (consecutive_highs < MAX_SAMPLE_ADJUST) ? -consecutive_highs : -MAX_SAMPLE_ADJUST;
		current_adjustment = ((buffered - upper_thresh) < MAX_SAMPLE_ADJUST) ? -(buffered - upper_thresh) : -MAX_SAMPLE_ADJUST;

#if LOG_SOUND
		fprintf(sound_log, "  (too high - adjusting to %d)\n", current_adjustment);
#endif
	}

	// otherwise, we're in the sweet spot
	else
	{
		// keep track of how many consecutive times we get this condition
		consecutive_lows = 0;
		consecutive_mids++;
		consecutive_highs = 0;

		// after 10 or so of these, revert back to no adjustment
		if (consecutive_mids > 10 && current_adjustment != 0)
		{
			current_adjustment = 0;

#if LOG_SOUND
			fprintf(sound_log, "  (%d consecutive_mids - adjusting to %d)\n", consecutive_mids, current_adjustment);
#endif
		}
	}
}

//============================================================
//	lock_buffer
//============================================================
static int lock_buffer(long offset, long size, void **buffer1, long *length1, void **buffer2, long *length2)
{
	volatile long pstart, pend, lstart, lend;

	if (!buf_locked)
	{
		if (video_config.throttle)
		{
			pstart = stream_playpos;
			pend = (pstart + SDL_XFER_SAMPLES);
			lstart = offset;
			lend = lstart+size;
			while (((pstart >= lstart) && (pstart <= lend)) ||
			       ((pend >= lstart) && (pend <= lend)))
			{
				#ifdef SDLMAME_WIN32
				Sleep(0);
				#else
				sleep(0);
				#endif

				pstart = stream_playpos;
				pend = pstart + SDL_XFER_SAMPLES;
			}
		}

		SDL_LockAudio();
		buf_locked++;
	}

	// init lengths
	*length1 = *length2 = 0;

	if ((offset + size) > stream_buffer_size)
	{
		// 2-piece case
		*length1 = stream_buffer_size - offset;
		*buffer1 = &stream_buffer[offset];
		*length2 = size - *length1;
		*buffer2 = stream_buffer;
	}
	else
	{
		// normal 1-piece case
		*length1 = size;
		*buffer1 = &stream_buffer[offset];
	}

	#if LOG_SOUND
//	fprintf(sound_log, "locking %ld bytes at offset %ld (total %d, playpos %d): len1 %ld len2 %ld\n",
//		size, offset, stream_buffer_size, stream_playpos, *length1, *length2);
	#endif

	return 0;
}

//============================================================
//	unlock_buffer
//============================================================
static void unlock_buffer(void)
{
	buf_locked--;
	if (!buf_locked)
		SDL_UnlockAudio();

	#if LOG_SOUND
//	fprintf(sound_log, "unlocking\n");
	#endif
}


//============================================================
//	copy_sample_data
//============================================================

static void copy_sample_data(UINT16 *data, int bytes_to_copy)
{
	void *buffer1, *buffer2 = (void *)NULL;
	long length1, length2;
	int cur_bytes;

	// attempt to lock the stream buffer
	if (lock_buffer(stream_buffer_in, bytes_to_copy, &buffer1, &length1, &buffer2, &length2) < 0)
	{
		buffer_underflows++;
		return;
	}

	// adjust the input pointer
	stream_buffer_in = (stream_buffer_in + bytes_to_copy) % stream_buffer_size;

	// copy the first chunk
	cur_bytes = (bytes_to_copy > length1) ? length1 : bytes_to_copy;
	memcpy(buffer1, data, cur_bytes);

	// adjust for the number of bytes
	bytes_to_copy -= cur_bytes;
	data = (UINT16 *)((UINT8 *)data + cur_bytes);

	// copy the second chunk
	if (bytes_to_copy != 0)
	{
		cur_bytes = (bytes_to_copy > length2) ? length2 : bytes_to_copy;
		memcpy(buffer2, data, cur_bytes);
	}

	// unlock
	unlock_buffer();
}


//============================================================
//	osd_update_audio_stream
//============================================================

int osd_update_audio_stream(INT16 *buffer)
{
	// if nothing to do, don't do it
	if (Machine->sample_rate != 0 && stream_buffer)
	{
		int original_bytes = bytes_in_stream_buffer();
		int input_bytes = samples_this_frame * nBlockAlign;
		int final_bytes;
		
		// update the sample adjustment
		update_sample_adjustment(original_bytes);
		
		// copy data into the sound buffer
		copy_sample_data((UINT16 *)buffer, input_bytes);
		
		// check for overflows
		final_bytes = bytes_in_stream_buffer();
		if (final_bytes < original_bytes)
			buffer_overflows++;
	}

	if( wavptr != NULL )
		wav_add_data_16( wavptr, buffer, samples_this_frame * 2 );

	// reset underflow/overflow tracking
	if (++total_frames == INGORE_UNDERFLOW_FRAMES)
		buffer_overflows = buffer_underflows = 0;

	// update underflow/overflow logging
#if (DISPLAY_UNDEROVERFLOW || LOG_SOUND)
{
	static int prev_overflows, prev_underflows;
	if (total_frames > INGORE_UNDERFLOW_FRAMES && (buffer_overflows != prev_overflows || buffer_underflows != prev_underflows))
	{
		prev_overflows = buffer_overflows;
		prev_underflows = buffer_underflows;
#if DISPLAY_UNDEROVERFLOW
		usrintf_showmessage("overflows=%d underflows=%d", buffer_overflows, buffer_underflows);
#endif
#if LOG_SOUND
//		fprintf(sound_log, "************************ overflows=%d underflows=%d\n", buffer_overflows, buffer_underflows);
#endif
	}
}
#endif

	// compute how many samples to generate next frame
	samples_left_over += samples_per_frame;
	samples_this_frame = (UINT32)samples_left_over;
	samples_left_over -= (double)samples_this_frame;
	
	samples_this_frame += current_adjustment;

	// return the samples to play this next frame
	return samples_this_frame;
}



//============================================================
//	osd_set_mastervolume
//============================================================

void osd_set_mastervolume(int _attenuation)
{
	// clamp the attenuation to 0-32 range
	if (_attenuation > 0)
		_attenuation = 0;
	if (_attenuation < -32)
		_attenuation = -32;
	attenuation = _attenuation;
}



//============================================================
//	osd_get_mastervolume
//============================================================

int osd_get_mastervolume(void)
{
	return attenuation;
}



//============================================================
//	osd_sound_enable
//============================================================

void osd_sound_enable(int enable_it)
{
	snd_enabled = enable_it;
}



//============================================================
//	osd_opl_control
//============================================================

void osd_opl_control(int chip, int reg)
{
	// noop - not supported
}



//============================================================
//	osd_opl_write
//============================================================

void osd_opl_write(int chip, int data)
{
	// noop - not supported
}

//============================================================
//	sdl_callback
//============================================================
void sdl_callback(void *userdata, Uint8 *stream, int len)
{
	int len1, len2;

	if ((stream_playpos+len) > stream_buffer_size)
	{
		len1 = stream_buffer_size - stream_playpos;
		len2 = len - len1;
	}
	else
	{
		len1 = len;
		len2 = 0;
	}

	if (snd_enabled)
	{
		memcpy(stream, stream_buffer + stream_playpos, len1);
		if (len2)
			memcpy(stream+len1, stream_buffer, len2);
	}
	else
	{
		memset(stream, 0, len);
	}

	// move the play cursor
	stream_playpos += len;
	stream_playpos %= stream_buffer_size;

	// always keep the write cursor just ahead of the play position
	stream_write_cursor = stream_playpos + len;
	stream_write_cursor %= stream_buffer_size;

	#if LOG_SOUND
//	fprintf(sound_log, "callback: xfer len1 %d len2 %d, playpos %d, wcursor %d\n",
//		len1, len2, stream_playpos, stream_write_cursor);
	#endif
}


//============================================================
//	sdl_init
//============================================================
static int sdl_init(void)
{
	int n_channels = 2;
	SDL_AudioSpec aspec;

	initialized_audio = 0;

	// set up the audio specs
	aspec.freq = Machine->sample_rate;
	aspec.format = AUDIO_S16SYS;	// keep endian independant 
	aspec.channels = n_channels;
	aspec.samples = SDL_XFER_SAMPLES;
	aspec.callback = sdl_callback;
	aspec.userdata = 0;

	nBlockAlign		= 16 * aspec.channels / 8;

	if (SDL_OpenAudio(&aspec, NULL) < 0)
		goto cant_start_audio;

	initialized_audio = 1;
	snd_enabled = 1;

	// compute the buffer sizes
	stream_buffer_size = ((UINT64)MAX_BUFFER_SIZE * (UINT64)Machine->sample_rate) / 44100;
	stream_buffer_size = (stream_buffer_size * nBlockAlign) / 4;
	stream_buffer_size = (stream_buffer_size * 30) / Machine->screen[0].refresh;
	stream_buffer_size = (stream_buffer_size / 1024) * 1024;

	// compute the upper/lower thresholds
	lower_thresh = audio_latency * stream_buffer_size / MAX_AUDIO_LATENCY;
	upper_thresh = (audio_latency + 1) * stream_buffer_size / MAX_AUDIO_LATENCY;

#if LOG_SOUND
	fprintf(sound_log, "stream_buffer_size = %d (max %d)\n", stream_buffer_size, MAX_BUFFER_SIZE);
	fprintf(sound_log, "lower_thresh = %d\n", lower_thresh);
	fprintf(sound_log, "upper_thresh = %d\n", upper_thresh);
#endif
//	printf("stream_buffer_size = %d (max %d)\n", stream_buffer_size, MAX_BUFFER_SIZE);
//	printf("lower_thresh = %d\n", lower_thresh);
//	printf("upper_thresh = %d\n", upper_thresh);

	// create the buffers
	if (sdl_create_buffers())
		goto cant_create_buffers;

	// start playing
	SDL_PauseAudio(0);
	return 0;

	// error handling
cant_create_buffers:
cant_start_audio:
	return 0;
}



//============================================================
//	sdl_kill
//============================================================

static void sdl_kill(void)
{
	if (initialized_audio)
		SDL_CloseAudio();
}



//============================================================
//	dsound_create_buffers
//============================================================

static int sdl_create_buffers(void)
{
//	printf("sdl_create_buffers: creating stream buffer of %x bytes\n", stream_buffer_size);
	stream_buffer = (INT8 *)malloc(stream_buffer_size);
	memset(stream_buffer, 0, stream_buffer_size);
	stream_playpos = 0;
	stream_write_cursor = SDL_XFER_SAMPLES*2;
	buf_locked = 0;
	return 0;
}



//============================================================
//	sdl_destroy_buffers
//============================================================

static void sdl_destroy_buffers(void)
{
	// release the buffer
	if (stream_buffer)
		free(stream_buffer);
	stream_buffer = NULL;
}

